from fetchfox import rest

BASE_URL = "https://js.cexplorer.io/api-static"


def get_asset_detail(fingerprint: str) -> dict:
    url = f"{BASE_URL}/asset/detail/{fingerprint}.json"

    response, status_code = rest.get(url)

    return response["data"]["tx"]



from fetchfox.blockchains.utils import check

ADDRESS_REGEX = r"^[A-Z2-7]{58}$"
ASSET_ID_REGEX = r"^[0-9]+$"
NF_DOMAIN_REGEX = r"^[a-z0-9]+([\-\.][a-z0-9]+)*\.algo$"


def is_address(string: str) -> bool:
    return check(ADDRESS_REGEX, string)


def is_asset_id(string: str) -> bool:
    return check(ASSET_ID_REGEX, string)


def is_nf_domain(string: str) -> bool:
    return check(NF_DOMAIN_REGEX, string)


def is_wallet(wallet: str) -> bool:
    return is_address(wallet) or is_nf_domain(wallet)



class InvalidCollectionIdException(ValueError):
    def __init__(self, string: str):
        super().__init__(f"'{string}' is not a valid collection id")


class InvalidAssetIdException(ValueError):
    def __init__(self, string: str):
        super().__init__(f"'{string}' is not a valid asset id")


class InvalidWalletException(ValueError):
    def __init__(self, string: str):
        super().__init__(f"'{string}' is not a valid wallet")



import json
import logging
from datetime import datetime
from typing import Iterable, Tuple, List

import pytz

from fetchfox.apis import bookio
from fetchfox.apis.evm import moralisio, ensideascom, openseaio
from fetchfox.blockchains.base import Blockchain
from fetchfox.constants.blockchains import ETHEREUM
from fetchfox.constants.marketplaces import OPENSEA_IO
from fetchfox.dtos import (
    AssetDTO,
    CampaignDTO,
    CampaignPricingDTO,
    FloorDTO,
    HoldingDTO,
    ListingDTO,
    SaleDTO,
)
from fetchfox.helpers import formatters
from . import utils
from .exceptions import (
    InvalidEvmAssetIdException,
    InvalidEvmCollectionIdException,
    InvalidEvmWalletException,
)

logger = logging.getLogger(__name__)


class Evm(Blockchain):
    def __init__(self, name: str, currency: str, logo: str, moralisio_api_key: str = None, openseaio_api_key: str = None):
        super().__init__(name, currency, logo)

        self.moralisio_api_key: str = moralisio_api_key
        self.openseaio_api_key: str = openseaio_api_key

    def check_asset_id(self, asset_id: str):
        if not utils.is_asset_id(asset_id):
            raise InvalidEvmAssetIdException(asset_id, self.name)

    def check_collection_id(self, collection_id: str):
        if not utils.is_address(collection_id):
            raise InvalidEvmCollectionIdException(collection_id, self.name)

    def check_wallet(self, wallet: str):
        if not utils.is_wallet(wallet):
            raise InvalidEvmWalletException(wallet, self.name)

    def get_wallet_name(self, wallet: str) -> str:
        if utils.is_ens_domain(wallet):
            return wallet

        if utils.is_address(wallet):
            return ensideascom.get_ens_domain(wallet)

        return None

    def resolve_wallet_name(self, wallet_name: str) -> str:
        if utils.is_ens_domain(wallet_name):
            return ensideascom.resolve_ens_domain(wallet_name)

        return None

    def get_asset(self, collection_id: str, asset_id: str, *args, **kwargs) -> AssetDTO:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        data = moralisio.get_asset_data(
            contract_address=collection_id,
            asset_id=asset_id,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
        )

        metadata = json.loads(data["metadata"])
        metadata["attributes"].update(metadata["extraAttributes"])

        return AssetDTO(
            collection_id=collection_id,
            asset_id=asset_id,
            metadata=metadata,
        )

    def get_assets(self, collection_id: str, fetch_metadata: bool = True, *args, **kwargs) -> Iterable[AssetDTO]:
        self.check_collection_id(collection_id)

        if fetch_metadata:
            asset_id = -1

            while True:
                try:
                    asset_id += 1

                    yield self.get_asset(
                        collection_id=collection_id,
                        asset_id=str(asset_id),
                    )
                except ValueError:
                    raise
                except:
                    break
        else:
            assets = moralisio.get_assets(
                contract_address=collection_id,
                blockchain=self.name,
                api_key=self.moralisio_api_key,
            )

            for asset_id in assets:
                yield AssetDTO(
                    collection_id=collection_id,
                    asset_id=asset_id,
                    metadata={},
                )

    def get_balance(self, wallet: str) -> Tuple[float, str]:
        self.check_wallet(wallet)

        if utils.is_ens_domain(wallet):
            wallet = ensideascom.resolve_ens_domain(wallet)

        balance = moralisio.get_balance(
            wallet,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
        )

        return balance, self.currency

    def get_campaigns(self, starts_after: datetime = None) -> Iterable[CampaignDTO]:
        def parse_pricing(pricing: dict) -> List[CampaignPricingDTO]:
            result = []

            if pricing.get("native_price"):
                result.append(
                    CampaignPricingDTO(
                        currency=self.currency,
                        amount=pricing["native_price"] / 10**18,
                    )
                )

            return result

        for campaign in bookio.get_campaigns():
            if not campaign.get("collection_id"):
                continue

            if campaign["blockchain"] != "evm":
                continue

            network = "mainnet" if self.name == ETHEREUM else self.name

            if campaign["network"] != network:
                continue

            collection_id = campaign["collection_id"]
            start_at = formatters.timestamp(campaign["start_at"])

            if starts_after is not None:
                if start_at < starts_after:
                    continue

            yield CampaignDTO(
                blockchain=self.name,
                parlamint_id=campaign["campaign_id"],
                collection_id=collection_id,
                name=campaign["name"],
                start_at=start_at,
                supply=campaign["total_deas"],
                limit=campaign["max_quantity"],
                mint_pricing=parse_pricing(
                    pricing=campaign.get("mint_price", {}),
                ),
                discount_pricing=parse_pricing(
                    pricing=campaign.get("discount_price", {}),
                ),
                bookstore_url=campaign["bookstore_url"],
                cover_url=campaign["cover_url"],
                explorer_url=self.explorer_url(
                    collection_id=collection_id,
                ),
            )

    def get_floor(self, collection_id: str, *args, **kwargs) -> FloorDTO:
        self.check_collection_id(collection_id)

        floor = None
        count = 0

        for listing in self.get_listings(collection_id):
            count += 1

            if floor is None:
                floor = listing
            elif listing.usd < floor.usd:
                floor = listing

        return FloorDTO(
            listing=floor,
            listing_count=count,
        )

    def get_holdings(self, wallet: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_wallet(wallet)

        if utils.is_ens_domain(wallet):
            wallet = ensideascom.resolve_ens_domain(wallet)

        holdings = moralisio.get_holdings(
            wallet,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
        )

        for token in holdings:
            contract_address = token["token_address"]
            asset_id = token["token_id"]
            quantity = int(token["amount"])

            yield HoldingDTO(
                collection_id=contract_address,
                asset_id=asset_id,
                address=wallet,
                quantity=quantity,
            )

    def get_listings(self, collection_id: str, slug: str = None, *args, **kwargs) -> Iterable[ListingDTO]:
        self.check_collection_id(collection_id)

        listings = openseaio.get_listings(
            collection_id,
            blockchain=self.name,
            api_key=self.openseaio_api_key,
            slug=slug,
        )

        for listing in listings:
            protocol_data = listing["protocol_data"]
            parameters = protocol_data["parameters"]

            price = listing["price"]["current"]

            asset_ids = []
            asset_names = []

            for offer in parameters["offer"]:
                if offer["token"].lower() != collection_id.lower():
                    continue

                asset_ids.append(offer["identifierOrCriteria"])
                asset_names.append("")

            if listing.get("eventTimestamp"):
                listed_at = datetime.utcfromtimestamp(parameters["startTime"])
            else:
                listed_at = datetime.now(tz=pytz.utc)

            marketplace_url = self.marketplace_url(
                collection_id=collection_id,
                asset_id=asset_ids[0],
            )

            yield ListingDTO(
                identifier=listing["order_hash"],
                collection_id=collection_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                listing_id=listing["order_hash"],
                marketplace=OPENSEA_IO,
                price=float(int(price["value"]) / 10 ** price["decimals"]),
                currency=price["currency"].replace("WETH", "ETH"),
                listed_at=listed_at,
                listed_by=parameters["offerer"],
                tx_hash=listing["order_hash"],
                marketplace_url=marketplace_url,
            )

    def get_owners(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        asset = moralisio.get_owner(
            collection_id,
            asset_id=asset_id,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
        )

        yield HoldingDTO(
            collection_id=asset["contract_address"],
            asset_id=asset["asset_id"],
            address=asset["address"],
            quantity=asset["amount"],
        )

    def get_sales(self, collection_id: str, slug: str = None, *args, **kwargs) -> Iterable[SaleDTO]:
        self.check_collection_id(collection_id)

        sales = openseaio.get_sales(
            collection_id,
            blockchain=self.name,
            api_key=self.openseaio_api_key,
            slug=slug,
        )

        for sale in sales:
            tx_hash = sale["transaction"]

            asset = sale["nft"]
            asset_id = asset["identifier"]
            asset_name = asset["name"]

            if sale.get("closing_date"):
                confirmed_at = datetime.fromtimestamp(
                    sale["closing_date"],
                    tz=pytz.utc,
                )
            else:
                confirmed_at = datetime.now(
                    tz=pytz.utc,
                )

            marketplace_url = self.marketplace_url(
                collection_id=collection_id,
                asset_id=asset_id,
            )

            explorer_url = self.explorer_url(
                tx_hash=tx_hash,
            )

            yield SaleDTO(
                identifier=f"{tx_hash}/{asset_id}",
                collection_id=collection_id,
                asset_ids=[asset_id],
                asset_names=[asset_name],
                tx_hash=tx_hash,
                marketplace=OPENSEA_IO,
                price=float(int(sale["payment"]["quantity"]) / 10 ** sale["payment"]["decimals"]),
                currency=sale["payment"]["symbol"].replace("WETH", "ETH"),
                confirmed_at=confirmed_at,
                sold_by=sale["seller"],
                bought_by=sale["buyer"],
                sale_id=sale["transaction"],
                marketplace_url=marketplace_url,
                explorer_url=explorer_url,
            )

    def get_snapshot(self, collection_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)

        assets = moralisio.get_owners(
            collection_id,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
        )

        for asset in assets:
            yield HoldingDTO(
                collection_id=asset["contract_address"],
                asset_id=asset["asset_id"],
                address=asset["address"],
                quantity=asset["amount"],
            )



from ..exceptions import (
    InvalidCollectionIdException,
    InvalidAssetIdException,
    InvalidWalletException,
)


class InvalidAlgorandCollectionIdException(InvalidCollectionIdException):
    def __init__(self, string: str):
        super().__init__(f"'{string}' is not a valid algorand creator address.")


class InvalidAlgorandAssetIdException(InvalidAssetIdException):
    def __init__(self, string: str):
        super().__init__(f"'{string}' is not a valid algorand asset id.")


class InvalidAlgorandWalletException(InvalidWalletException):
    def __init__(self, string: str):
        super().__init__(f"'{string}' is not a valid algorand address or non-fungible domain.")



import logging
import time

from cachetools.func import ttl_cache

from fetchfox import rest
from fetchfox.constants.currencies import ALGO, ADA, BOOK, ETH, MATIC

IDS = {
    ALGO: "algorand",
    ADA: "cardano",
    BOOK: "book-2",
    ETH: "ethereum",
    MATIC: "matic-network",
}

logger = logging.getLogger(__name__)


@ttl_cache(ttl=60 * 60)
def usd(currency: str):
    time.sleep(5)

    currency = currency.strip().upper()

    id = IDS[currency]

    logger.info("fetching exchange for %s (%s)", currency, id)

    response, status_code = rest.get(
        url="https://api.coingecko.com/api/v3/simple/price",
        params={
            "ids": id,
            "vs_currencies": "usd",
        },
    )

    return response[id]["usd"]


@ttl_cache(ttl=60 * 60)
def ath_usd(currency: str):
    time.sleep(5)

    currency = currency.strip().upper()

    id = IDS[currency]

    logger.info("fetching ath for %s (%s)", currency, id)

    response, status_code = rest.get(
        url=f"https://api.coingecko.com/api/v3/coins/{id}",
    )

    return response["market_data"]["ath"]["usd"]



from ..exceptions import (
    InvalidCollectionIdException,
    InvalidAssetIdException,
    InvalidWalletException,
)


class InvalidCardanoCollectionIdException(InvalidCollectionIdException):
    def __init__(self, string: str):
        super().__init__(f"'{string}' is not a valid cardano policy id.")


class InvalidCardanoAssetIdException(InvalidAssetIdException):
    def __init__(self, string: str):
        super().__init__(f"'{string}' is not a valid cardano asset id.")


class InvalidCardanoWalletException(InvalidWalletException):
    def __init__(self, string: str):
        super().__init__(f"'{string}' is not a valid cardano stake key, address or ada handle.")



import logging
from functools import lru_cache
from typing import Tuple, Iterable

from fetchfox import rest
from fetchfox.checks import check_str
from fetchfox.constants.blockchains import ETHEREUM, POLYGON

BASE_URL = "https://api.opensea.io"

BLOCKCHAINS = {
    ETHEREUM: "ethereum",
    POLYGON: "matic",
}

logger = logging.getLogger(__name__)


def get(service: str, api_key: str, params: dict = None, version: int = 2, check: str = None) -> Tuple[dict, int]:
    check_str(api_key, "openseaio.api_key")

    if version == 1:
        url = f"{BASE_URL}/api/v{version}/{service}"
    else:
        url = f"{BASE_URL}/v{version}/{service}"

    return rest.get(
        url=url,
        headers={
            "X-API-KEY": api_key,
        },
        params=params or {},
        sleep=2.5,
        check=check,
    )


@lru_cache(maxsize=None)
def get_slug(contract_address: str, blockchain: str, api_key: str) -> str:
    check_str(contract_address, "openseaio.contract_address")
    check_str(blockchain, "openseaio.blockchain")

    contract_address = contract_address.strip().lower()
    blockchain: str = BLOCKCHAINS.get(blockchain, blockchain)

    logger.info("fetching slug for %s (%s)", contract_address, blockchain)

    response, status_code = get(
        f"chain/{blockchain}/contract/{contract_address}",
        api_key=api_key,
        check="collection",
    )

    return response["collection"]


def get_sales(contract_address: str, blockchain: str, api_key: str, slug: str = None) -> Iterable[dict]:
    check_str(contract_address, "openseaio.contract_address")
    check_str(blockchain, "openseaio.blockchain")

    contract_address = contract_address.strip().lower()

    if not slug:
        slug = get_slug(contract_address, blockchain, api_key=api_key)

    cursor = ""

    while True:
        response, status_code = get(
            f"events/collection/{slug}",
            params={
                "event_type": "sale",
                "next": cursor,
            },
            api_key=api_key,
        )

        for event in response["asset_events"]:
            if event["event_type"] != "sale":
                continue

            yield event

        cursor = response.get("next")

        if not cursor:
            break


def get_listings(contract_address: str, blockchain: str, api_key: str, slug: str = None) -> Iterable[dict]:
    check_str(contract_address, "openseaio.contract_address")
    check_str(blockchain, "openseaio.blockchain")

    contract_address = contract_address.strip().lower()

    if not slug:
        slug = get_slug(contract_address, blockchain, api_key=api_key)

    cursor = ""

    while True:
        response, status_code = get(
            f"listings/collection/{slug}/all",
            params={
                "next": cursor,
            },
            api_key=api_key,
        )

        yield from response.get("listings", [])

        cursor = response.get("next")

        if not cursor:
            break



from typing import Dict


class StatsDTO:
    def __init__(
        self,
        ada: float,
        book: float,
        daily_txs: int,
        daily_buys: int,
        daily_sales: int,
        daily_volume: float,
        price_change_hour: float,
        price_change_day: float,
        price_change_week: float,
        price_change_month: float,
    ):
        self.ada: float = ada
        self.book: float = book
        self.daily_txs: int = daily_txs
        self.daily_buys: int = daily_buys
        self.daily_sales: int = daily_sales
        self.daily_volume: float = daily_volume
        self.price_change_hour: float = price_change_hour
        self.price_change_day: float = price_change_day
        self.price_change_week: float = price_change_week
        self.price_change_month: float = price_change_month

    def __repr__(self):
        return f"{round(self.ada)} ADA - {round(self.book)} BOOK"


class PairStatsDTO:
    def __init__(self, stats: StatsDTO):
        self.stats: StatsDTO = stats
        self.pools: Dict[StatsDTO] = {}

    def add_pool(self, name: str, stats: StatsDTO):
        self.pools[name.lower()] = stats

    def __repr__(self):
        return f"{self.stats} ({len(self.pools)} pools)"



from typing import Dict

from fetchfox.constants.specials import GHOST_BIBLE, STRABBLE_BIBLE


class AssetDTO:
    def __init__(self, collection_id: str, asset_id: str, metadata: dict):
        self.collection_id: str = collection_id
        self.asset_id: str = asset_id
        self.metadata: dict = metadata

    @property
    def name(self) -> str:
        return self.metadata["name"]

    @property
    def title(self) -> str:
        if "Book Title" in self.metadata:
            return self.metadata["Book Title"][0]

        return self.metadata["name"].split(" #")[0]

    @property
    def number(self) -> str:
        try:
            return int(self.metadata["name"].split(" #")[-1])
        except:
            return None

    @property
    def quantity(self) -> int:
        return int(self.metadata.get("quantity", "1"))

    @property
    def cover_theme(self) -> str:
        attributes = self.metadata.get("attributes") or self.metadata.get("properties") or {}

        if not attributes:
            return "none"

        if "Cover Theme" in attributes:
            return attributes["Cover Theme"].split(" / ")[-1]

        if "Edition" in attributes:
            return attributes["Edition"]

        trait_count = 0

        for attribute in attributes.values():
            if isinstance(attribute, str):
                trait_count += 1
            elif isinstance(attribute, list):
                trait_count += len(attribute)

        return f"Traits: {trait_count}"

    @property
    def cover_variation(self) -> str:
        attributes = self.metadata.get("attributes") or self.metadata.get("properties") or {}

        if not attributes:
            return None

        try:
            return int(attributes["Variation"].split(" / ")[-1])
        except:
            return None

    @property
    def files(self) -> Dict[str, str]:
        return {item["name"]: item["src"] for item in self.metadata.get("files", [])}

    def image_url(self, https: bool = False, highres: bool = False) -> str:
        url = None

        if highres:
            url = self.files.get("High-Res Cover Image")

        if not url:
            url = self.metadata.get("image") or self.metadata.get("media_url")

        if https and url:
            url = url.replace("ipfs://", "https://ipfs.io/ipfs/")

        return url

    @property
    def special(self) -> str:
        attributes = self.metadata.get("attributes") or self.metadata.get("properties") or {}

        if not attributes:
            return None

        if self.metadata["name"].startswith("Gutenberg Bible"):
            dots = attributes.get("Dots", [])
            black_knot = attributes.get("Black_Knot", [])

            specials = []

            if "Dots_Middle" in dots and "Dots_ADA" not in dots:
                specials.append(GHOST_BIBLE)

            if "Black_Knot_Outer_Ring_Stroke" in black_knot and "Black_Knot_Inner_Ring_Stroke" not in black_knot:
                specials.append(STRABBLE_BIBLE)

            if specials:
                return " - ".join(specials)

        return attributes.get("Special")

    @property
    def emoji(self) -> str:
        emojis = ""

        if GHOST_BIBLE in (self.special or ""):
            emojis += "👻"

        if STRABBLE_BIBLE in (self.special or ""):
            emojis += "🖌"

        if self.special in ["Bonus story in book", "Bonus chapter in book"]:
            emojis += "📖"

        return emojis if len(emojis) > 0 else None

    def __repr__(self) -> str:
        return f"{self.title} {self.number} [{self.cover_theme} / {self.cover_variation}]"



from datetime import datetime, timedelta
from typing import List
from urllib.parse import urlencode

import pytz


class CampaignPricingDTO:
    def __init__(self, currency: str, amount: float):
        self.currency: str = currency
        self.amount: float = amount

    def __repr__(self) -> str:
        return f"{self.amount:0.2f} {self.currency}"


class CampaignDTO:
    def __init__(
        self,
        blockchain: str,
        parlamint_id: str,
        collection_id: str,
        name: str,
        start_at: datetime,
        supply: int,
        limit: int,
        mint_pricing: List[CampaignPricingDTO],
        discount_pricing: List[CampaignPricingDTO],
        bookstore_url: str,
        cover_url: str,
        explorer_url: str,
    ):
        self.blockchain: str = blockchain

        self.parlamint_id: str = parlamint_id.lower()
        self.collection_id: str = collection_id.lower()

        self.name: str = name
        self.start_at: datetime = start_at
        self.supply: int = supply
        self.limit: int = limit
        self.mint_pricing: List[CampaignPricingDTO] = mint_pricing
        self.discount_pricing: List[CampaignPricingDTO] = discount_pricing

        self.bookstore_url: str = bookstore_url
        self.cover_url: str = cover_url
        self.explorer_url: str = explorer_url

    @property
    def parlamint_url(self) -> str:
        return f"https://app.book.io/parlamint-v2/{self.parlamint_id}"

    @property
    def new(self) -> bool:
        now = datetime.utcnow().replace(tzinfo=pytz.UTC)

        return self.start_at > now

    @property
    def timedelta(self) -> str:
        if not self.start_at:
            return None

        now = datetime.utcnow().replace(tzinfo=pytz.UTC)

        if self.start_at > now:
            td = self.start_at - now
        else:
            td = now - self.start_at

        days = td.days
        hours = td.seconds // 3600
        minutes = (td.seconds // 60) % 60

        if days == hours == minutes == 0:
            return "NOW!!"

        items = []

        if days:
            items += [
                "{days} day{s}".format(
                    days=abs(days),
                    s="" if days == 1 else "s",
                )
            ]

        if days < 2:
            if days and (hours or minutes):
                items += ["and"]

            if hours or minutes:
                items += [
                    "{hours:02d}:{minutes:02d} hs".format(
                        hours=hours,
                        minutes=minutes,
                    )
                ]

        if self.start_at < now:
            items += ["ago"]

        return " ".join(items)

    def googlecalendar_url(self) -> str:  # https://www.labnol.org/calendar/
        start = self.start_at
        end = start + timedelta(minutes=30)

        start_str = "{year}{month:02}{day:02}T{hour:02}{minute:02}00Z".format(
            year=start.year,
            month=start.month,
            day=start.day,
            hour=start.hour,
            minute=start.minute,
        )

        end_str = "{year}{month:02}{day:02}T{hour:02}{minute:02}00Z".format(
            year=end.year,
            month=end.month,
            day=end.day,
            hour=end.hour,
            minute=end.minute,
        )

        params = {
            "action": "TEMPLATE",
            "location": self.parlamint_url,
            "text": f"BOOK.IO Mint: {self.name}",
            "dates": f"{start_str}/{end_str}",
        }

        return f"https://calendar.google.com/calendar/render?{urlencode(params)}"

    def __repr__(self):
        return f"{self.name} ({self.collection_id})"
